/*
 * VeriSilicon sample for encoding with VPE suppprted avcodec API.
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <sys/time.h>
#include <unistd.h>
#include <semaphore.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>

#include <libavcodec/avcodec.h>
#include <libavcodec/internal.h>
#include <libavformat/avformat.h>
#include <libavfilter/buffersink.h>
#include <libavfilter/buffersrc.h>
#include <libavutil/opt.h>
#include <libavutil/log.h>
#include <libavutil/pixdesc.h>
#include <libavutil/hwcontext.h>
#include <libavutil/imgutils.h>

static const int avloglevel = AV_LOG_INFO;
static const char *device   = "/dev/ama_transcoder0";
static char *enc_codec;
static char *output_file;

static const char *device_option_string = "priority=vod,vpeloglevel=6";
static const char *filter_descr = "vpe_pp";
static const char *enc_option_string = "preset=fast,b=10000000";
static const char *enc_params_key = "enc_params";
static const char *enc_params_value = "gop_size=1";

static int width, height;
static FILE *fp = NULL;
static FILE *fq = NULL;

static AVBufferRef *hw_device_ctx = NULL;

static AVCodecContext *enc_ctx;
AVFilterContext *buffersink_ctx;
AVFilterContext *buffersrc_ctx;
AVFilterGraph *filter_graph;
enum AVPixelFormat pix_fmt;

static int encode_frame(AVFrame *frame, AVPacket *pkt)
{
    int ret;

    /* send the frame to the encoder */
    if (frame)
        av_log(NULL, AV_LOG_DEBUG, "Send frame %3"PRId64"\n", frame->pts);

    ret = avcodec_send_frame(enc_ctx, frame);
    if (ret < 0) {
        fprintf(stderr, "Error sending a frame for encoding\n");
        return ret;
    }

    while (ret >= 0) {
        /* get encoded packet */
        ret = avcodec_receive_packet(enc_ctx, pkt);
        if (ret == AVERROR(EAGAIN) || ret == AVERROR_EOF)
            return 0;
        else if (ret < 0) {
            fprintf(stderr, "Error during encoding\n");
            return ret;
        }

        av_log(NULL, AV_LOG_DEBUG, "Write packet %3"PRId64" (size=%5d)\n", pkt->pts, pkt->size);

        if (fp) {
            fwrite(pkt->data, 1, pkt->size, fp);
            fflush(fp);
        }
        av_packet_unref(pkt);
    }

    return 0;
}

static int encode_flush(AVPacket *pkt)
{
    int ret;

    for (;;) {
        while ((ret = avcodec_receive_packet(enc_ctx, pkt)) == AVERROR(EAGAIN)) {
            ret = avcodec_send_frame(enc_ctx, NULL);
            if (ret < 0) {
                av_log(NULL, AV_LOG_FATAL, "encoding failed: %s\n", av_err2str(ret));
                return -1;
            }
        }
        if (ret < 0 && ret != AVERROR_EOF) {
            av_log(NULL, AV_LOG_FATAL, "encoding failed: %s\n", av_err2str(ret));
            return -1;
        }
        if (fp && ret != AVERROR_EOF) {
            fwrite(pkt->data, 1, pkt->size, fp);
            fflush(fp);
        }
        av_packet_unref(pkt);
        if (ret == AVERROR_EOF) {
            break;
        }
    }
    return 0;
}

static int init_vpe_device(void)
{
    int ret;
    AVDictionary *opts = NULL;

    av_dict_parse_string(&opts, device_option_string, "=", ",", 0);
    ret = av_hwdevice_ctx_create(&hw_device_ctx, AV_HWDEVICE_TYPE_AMA, device,
                                 opts, 0);
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR,
               "Failed to create a VPE device. Error code: %s\n",
               av_err2str(ret));
        return -1;
    }
    return 0;
}

static int init_filters(const char *filters_spec)
{
    char args[512];
    int ret                         = 0;
    const AVFilter *buffersrc       = NULL;
    const AVFilter *buffersink      = NULL;
    AVFilterInOut *outputs          = avfilter_inout_alloc();
    AVFilterInOut *inputs           = avfilter_inout_alloc();
    AVFilterGraph *filter_graph     = avfilter_graph_alloc();

    AVRational time_base = (AVRational){1, 25};
    AVRational sample_aspect_ratio = (AVRational){1, 1};
    int i;

    if (!outputs || !inputs || !filter_graph) {
        ret = AVERROR(ENOMEM);
        goto end;
    }

    av_log(NULL, AV_LOG_DEBUG, "filters_spec '%s' for stream\n", filters_spec);

    buffersrc  = avfilter_get_by_name("buffer");
    buffersink = avfilter_get_by_name("buffersink");
    if (!buffersrc || !buffersink) {
        av_log(NULL, AV_LOG_ERROR,
               "filtering source or sink element not found\n");
        ret = AVERROR_UNKNOWN;
        goto end;
    }

    snprintf(args, sizeof(args),
             "video_size=%dx%d:pix_fmt=%d:time_base=%d/%d:pixel_aspect=%d/"
             "%d",
             width, height, pix_fmt,
             time_base.num, time_base.den,
             sample_aspect_ratio.num,
             sample_aspect_ratio.den);
    printf("args %s\n", args);

    ret = avfilter_graph_create_filter(&buffersrc_ctx, buffersrc, "in",
                                       args, NULL, filter_graph);
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR, "Cannot create buffer source\n");
        goto end;
    }
    ret = avfilter_graph_create_filter(&buffersink_ctx, buffersink, "out",
                                       NULL, NULL, filter_graph);
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR, "Cannot create buffer sink\n");
        goto end;
    }

    /* Endpoints for the filter graph. */
    outputs->name       = av_strdup("in");
    outputs->filter_ctx = buffersrc_ctx;
    outputs->pad_idx    = 0;
    outputs->next       = NULL;

    inputs->name       = av_strdup("out");
    inputs->filter_ctx = buffersink_ctx;
    inputs->pad_idx    = 0;
    inputs->next       = NULL;

    if (!outputs->name || !inputs->name) {
        ret = AVERROR(ENOMEM);
        goto end;
    }

    av_log(NULL, AV_LOG_DEBUG, "to avfilter_graph_parse_ptr\n");

    if ((ret = avfilter_graph_parse_ptr(filter_graph, filters_spec, &inputs,
                                        &outputs, NULL)) < 0)
        goto end;

    av_log(NULL, AV_LOG_DEBUG, "to set hwdevice to filter\n");

    if (hw_device_ctx) {
        for (i = 0; i < filter_graph->nb_filters; i++) {
            filter_graph->filters[i]->hw_device_ctx =
                av_buffer_ref(hw_device_ctx);
            if (!filter_graph->filters[i]->hw_device_ctx) {
                ret = AVERROR(ENOMEM);
                goto end;
            }
        }
    }

    av_log(NULL, AV_LOG_DEBUG, "to avfilter_graph_config\n");

    if ((ret = avfilter_graph_config(filter_graph, NULL)) < 0)
        goto end;

end:
    avfilter_inout_free(&inputs);
    avfilter_inout_free(&outputs);

    return ret;
}

static int open_output_file(void)
{
    AVCodec *enc;
    int ret;
    unsigned int i;
    AVBufferRef *hwfrm;
    AVDictionary *opts = NULL;

    av_log(NULL, AV_LOG_TRACE, "%s(%d)\n", __FUNCTION__, __LINE__);

    enc = avcodec_find_encoder_by_name(enc_codec);
    if (!enc) {
        av_log(NULL, AV_LOG_FATAL, "Necessary encoder not found\n");
        av_log(NULL, AV_LOG_INFO, "Supported encoder: h264enc_vpe, hevcenc_vpe, vp9enc_vpe\n");
        return AVERROR_INVALIDDATA;
    }
    enc_ctx = avcodec_alloc_context3(enc);
    if (!enc_ctx) {
        av_log(NULL, AV_LOG_FATAL,
                "Failed to allocate the encoder context\n");
        return AVERROR(ENOMEM);
    }

    /* In this example, we transcode to same properties (picture size,
     * sample rate etc.). These properties can be changed for output
     * streams easily using filters */
    enc_ctx->height = buffersink_ctx->inputs[0]->h;
    enc_ctx->width = buffersink_ctx->inputs[0]->w;
    enc_ctx->sample_aspect_ratio =
        buffersink_ctx->inputs[0]->sample_aspect_ratio;
    /* take first format from list of supported formats */
    enc_ctx->pix_fmt = AV_PIX_FMT_VPE;
    /* video time_base can be set to whatever is handy and supported
     * by encoder */
    enc_ctx->time_base = buffersink_ctx->inputs[0]->time_base;
    enc_ctx->framerate = (AVRational){25, 1};

    av_dict_parse_string(&opts, enc_option_string, "=", ",", 0);
    av_dict_set(&opts, enc_params_key, enc_params_value, 0);
    /* set hw device for encoder */
    enc_ctx->hw_device_ctx =
        av_buffer_ref(buffersink_ctx->hw_device_ctx);
    if (!enc_ctx->hw_device_ctx) {
        av_log(NULL, AV_LOG_ERROR,
              "A hardware device reference create failed.\n");
        return AVERROR(ENOMEM);
    }

    /* set hw frame for encoder */
    hwfrm =
        av_buffersink_get_hw_frames_ctx(buffersink_ctx);
        av_log(NULL, AV_LOG_DEBUG, "hwfrm = %p\n", hwfrm);

    if (hwfrm &&
        ((AVHWFramesContext *)hwfrm->data)->format ==
        av_buffersink_get_format(buffersink_ctx)) {
        av_log(NULL, AV_LOG_DEBUG, "set enc hwframe\n");
        enc_ctx->hw_frames_ctx = av_buffer_ref(hwfrm);
        if (!enc_ctx->hw_frames_ctx)
            return AVERROR(ENOMEM);
    }

    /* Third parameter can be used to pass settings to encoder */
    ret = avcodec_open2(enc_ctx, enc, &opts);
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR,
               "Cannot open video encoder for stream #%u\n", i);
        return ret;
    }

    return 0;
}

/*
 * This function is used to push raw data to AVFrame;
 * If using data read from input file, write file_name in argv[6]
 * If using data read from memory, write valid memory address in argv[6]
 */
static int feed_input_data(char *data, AVFrame *frame, int size)
{
    int err;

    if (!fq) {
        if (!(fq = fopen(data, "rb"))) {
            printf("Input is not file, will use memory\n");
        }
    }

    if (fq) {
        if (pix_fmt == AV_PIX_FMT_NV12) {
            if ((err = fread((uint8_t*)(frame->data[0]), 1, size, fq)) <= 0)
                return 0;
            if ((err = fread((uint8_t*)(frame->data[1]), size/2, 1, fq)) <= 0)
                return 0;
        } else if (pix_fmt == AV_PIX_FMT_RGBA ||
                   pix_fmt == AV_PIX_FMT_ARGB ||
                   pix_fmt == AV_PIX_FMT_ABGR ||
                   pix_fmt == AV_PIX_FMT_BGRA) {
            if ((err = fread((uint8_t*)(frame->data[0]), 1, size*4, fq)) <= 0)
                return 0;
        } else if (pix_fmt == AV_PIX_FMT_RGB24 ||
                   pix_fmt == AV_PIX_FMT_BGR24) {
            if ((err = fread((uint8_t*)(frame->data[0]), 1, size*3, fq)) <= 0)
                return 0;
        }
    } else if (data) {
        if (pix_fmt == AV_PIX_FMT_NV12) {
            memcpy(frame->data[0], data, size);
            memcpy(frame->data[1], data+size, size/2);
        } else if (pix_fmt == AV_PIX_FMT_RGBA ||
                   pix_fmt == AV_PIX_FMT_ARGB ||
                   pix_fmt == AV_PIX_FMT_ABGR ||
                   pix_fmt == AV_PIX_FMT_BGRA) {
            memcpy(frame->data[0], data, size*4);
        } else if (pix_fmt == AV_PIX_FMT_RGB24 ||
                   pix_fmt == AV_PIX_FMT_BGR24) {
            memcpy(frame->data[0], data, size*3);
        }
    }

    return 1;
}

int main(int argc, char *argv[])
{
    int size;
    AVFrame *sw_frame = NULL;
    AVFrame *filt_frame;
    AVPacket *enc_pkt;
    int ret;

    if (argc < 6) {
        fprintf(stderr, "Usage: %s <width> <height> <pix_fmt> <vpe encoding codec name> <output file> <input file>\n", argv[0]);
        return -1;
    }

    av_log_set_level(avloglevel);

    ret = init_vpe_device();
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR, "Failed to init a VPE device\n");
        return -1;
    }
    enc_codec = malloc(128);
    strcpy(enc_codec, argv[4]);

    if (argc <= 5) {
        output_file = NULL;
    } else {
        output_file = malloc(128);
        strcpy(output_file, argv[5]);
        fp = fopen(output_file, "wb");
        if (fp == NULL) {
            av_log(NULL, AV_LOG_ERROR, "Failed to open output file\n");
            return -1;
        }
    }

    filt_frame = av_frame_alloc();
    if (!filt_frame) {
        perror("Could not allocate frame");
        goto end;
    }

    width  = atoi(argv[1]);
    height = atoi(argv[2]);
    size   = width * height;

    if ((pix_fmt = av_get_pix_fmt(argv[3])) == AV_PIX_FMT_NONE) {
        printf("No such pixel format: %s.\n", argv[3]);
        goto end;
    }

    if ((ret = init_filters(filter_descr)) < 0)
        goto end;

    ret = open_output_file();
    if (ret < 0)
        goto end;

    /* alloc packet for encoder output */
    enc_pkt = av_packet_alloc();
    if (!enc_pkt)
        goto end;

    while (1) {
        if (!(sw_frame = av_frame_alloc())) {
            ret = AVERROR(ENOMEM);
            goto end;
        }
        /* read data into software frame, and transfer them into hw frame */
        sw_frame->width  = width;
        sw_frame->height = height;
        sw_frame->format = pix_fmt;

        if ((ret = av_frame_get_buffer(sw_frame, 0)) < 0)
            goto end;

        // push raw data to PP
        if (0 == feed_input_data(argv[6], sw_frame, size))
            break;
        /* push the raw data frame into the filtergraph */
        if (av_buffersrc_add_frame_flags(buffersrc_ctx, sw_frame, AV_BUFFERSRC_FLAG_KEEP_REF) < 0) {
            av_log(NULL, AV_LOG_ERROR, "Error while feeding the filtergraph\n");
            break;
        }

        while (1) {
            /* pull filtered frames from the filtergraph */
            ret = av_buffersink_get_frame(buffersink_ctx, filt_frame);
            if (ret == AVERROR(EAGAIN) || ret == AVERROR_EOF)
                break;
            if (ret < 0)
                goto end;

            /* push filtered frames to encoder and get encoded packet */
            if (encode_frame(filt_frame, enc_pkt) < 0)
                goto end;
            av_frame_unref(filt_frame);
        }

        av_frame_free(&sw_frame);
    }

    /* flush encoder */
    ret = encode_flush(enc_pkt);

end:
    av_packet_free(&enc_pkt);
    av_frame_free(&sw_frame);
    av_frame_free(&filt_frame);
    avcodec_free_context(&enc_ctx);
    avfilter_graph_free(&filter_graph);
    av_buffer_unref(&hw_device_ctx);

    free(enc_codec);
    if (fp) {
        fclose(fp);
    }
    if (fq) {
        fclose(fq);
    }

    return ret ? 1 : 0;
}