/*
 * VeriSilicon VPE GC620 process interface
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "libavutil/avstring.h"
#include "libavutil/imgutils.h"
#include "libavutil/opt.h"
#include "libavutil/parseutils.h"
#include "libavutil/pixdesc.h"

#include "avfilter.h"
#include "drawutils.h"
#include "formats.h"
#include "internal.h"
#include "framesync.h"
#include "video.h"

#include <libavutil/hwcontext.h>
#include <libavutil/hwcontext_vpe.h>
#include "vpi_types.h"
#include "vpi_api.h"

enum { R, G, B, A };
typedef struct VpeGC620Context {
    const AVClass *class;
    const AVPixFmtDescriptor *desc;
    unsigned x, y;
    int nb_inputs;
    FFFrameSync fs;
    int p_type;
    char *p_type_str;
    int core_id;

    /* for stack */
    AVFrame **frames;
    char *layout;
    int shortest;

    /* for drawbox */
    int thickness;
    char *color_str;
    uint8_t yuv_color[4];
    uint8_t rgb_color[4];

    /* for overlay */
    int overlay_x;
    int overlay_y;

    VpiProcInitParams vpe_gc620_params;
    VpiProcHandle *vpe_gc620_handle;

    AVBufferRef *hw_frame_ctx;
    AVBufferRef *hwdevice;

} VpeGC620Context;

static int vpe_init_hwdev(AVFilterContext *ctx)
{
    VpeGC620Context *s = ctx->priv;

    if (!s->hwdevice) {
        if (ctx->hw_device_ctx) {
            s->hwdevice = av_buffer_ref(ctx->hw_device_ctx);
            if (!s->hwdevice) {
                return AVERROR(ENOMEM);
            }
        } else {
            return AVERROR_UNKNOWN;
        }
    }

    return 0;
}

static int vpe_check_constrains(AVFilterContext *ctx)
{
    VpeGC620Context *s = ctx->priv;
    AVHWFramesConstraints *constrains;
    AVFilterLink *inlink    = ctx->inputs[0];
    AVVpeHWConfig *hwconfig = NULL;
    int ret = 0;

    hwconfig = av_hwdevice_hwconfig_alloc(s->hwdevice);
    if (!hwconfig) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    hwconfig->id = VPE_HW_TILE;

    constrains = av_hwdevice_get_hwframe_constraints(s->hwdevice, hwconfig);
    if (!constrains) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    if (inlink->w < constrains->min_width ||
        inlink->h < constrains->min_height ||
        inlink->w > constrains->max_width ||
        inlink->h > constrains->max_height) {
        av_log(ctx, AV_LOG_ERROR,
               "size %dx%d is not supported. "
               "constrains: width %d-%d height %d-%d.\n",
               inlink->w, inlink->h, constrains->min_width,
               constrains->max_width, constrains->min_height,
               constrains->max_height);
        ret = AVERROR(EINVAL);
        goto error;
    }

error:
    if (constrains)
        av_hwframe_constraints_free(&constrains);
    if (hwconfig)
        av_freep(&hwconfig);

    return ret;
}

static int vpe_init_hwctx(AVFilterContext *ctx)
{
    VpeGC620Context *s        = ctx->priv;
    VpiProcInitParams *param = &s->vpe_gc620_params;
    AVHWFramesContext *hw_frame_ctx;
    AVHWDeviceContext *device_ctx;
    AVVpeDeviceContext *device_hwctx;
    AVVpeFramesContext *frame_hwctx;
    AVFilterLink *inlink = ctx->inputs[0];
    int ret;

    /* hw device & frame init */
    ret = vpe_init_hwdev(ctx);
    if (ret < 0) {
        return ret;
    }

    if (inlink->hw_frames_ctx) {
        av_log(ctx, AV_LOG_DEBUG, "%s(%d)\n", __FUNCTION__, __LINE__);
        s->hw_frame_ctx = av_buffer_ref(inlink->hw_frames_ctx);
        if (!s->hw_frame_ctx) {
            ret = AVERROR(ENOMEM);
            return ret;
        }

        hw_frame_ctx = (AVHWFramesContext *)s->hw_frame_ctx->data;
    } else {
        av_log(ctx, AV_LOG_ERROR, "%s(%d)\n", __FUNCTION__, __LINE__);
        ret = AVERROR_UNKNOWN;
        return ret;
    }

    ret = vpe_check_constrains(ctx);
    if (ret < 0) {
        return ret;
    }

    device_ctx   = hw_frame_ctx->device_ctx;
    device_hwctx = device_ctx->hwctx;
    frame_hwctx  = hw_frame_ctx->hwctx;

    if (!inlink->hw_frames_ctx) {
        inlink->hw_frames_ctx = av_buffer_ref(s->hw_frame_ctx);
        if (!inlink->hw_frames_ctx) {
            ret = AVERROR(ENOMEM);
            return ret;
        }
    }

    param->dev_ctx = device_hwctx->vpe_dev_ctx;
    param->mem_ctx = frame_hwctx->vpe_mem_ctx;
    av_log(ctx, AV_LOG_DEBUG, "%s: vpe_mem_ctx = %p.\n", __func__,
           frame_hwctx->vpe_mem_ctx);

    return 0;
}

static int query_formats(AVFilterContext *ctx)
{
    VpeGC620Context *s                  = ctx->priv;
    AVVpeHWConfig *hwconfig            = NULL;
    AVHWFramesConstraints *constraints = NULL;
    AVFilterFormats *in_fmts;
    AVFilterFormats *out_fmts;
    int ret;
    int i;

    ret = vpe_init_hwdev(ctx);
    if (ret < 0) {
        goto error;
    }

    hwconfig = av_hwdevice_hwconfig_alloc(s->hwdevice);
    if (!hwconfig) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    hwconfig->id = VPE_HW_TILE;
    av_log(ctx, AV_LOG_DEBUG, "hwconfig->id = %d\n", hwconfig->id);

    constraints = av_hwdevice_get_hwframe_constraints(s->hwdevice, hwconfig);
    if (!constraints) {
        av_log(ctx, AV_LOG_ERROR, "failed to get hwframe constrains.\n");
        ret = AVERROR(ENOMEM);
        goto error;
    }

    av_log(ctx, AV_LOG_DEBUG, "%s(%d)\n", __FUNCTION__, __LINE__);

    out_fmts = ff_make_format_list(constraints->valid_hw_formats);
    ret      = ff_formats_ref(out_fmts, &ctx->outputs[0]->incfg.formats);
    av_log(ctx, AV_LOG_DEBUG, "%s(%d) ret = %d\n", __FUNCTION__, __LINE__, ret);
    if (ret < 0)
        goto error;

    in_fmts = ff_make_format_list(constraints->valid_hw_formats);
    i       = 0;
    while (constraints->valid_sw_formats[i] != AV_PIX_FMT_NONE) {
        ret = ff_add_format(&in_fmts, constraints->valid_sw_formats[i]);
        av_log(ctx, AV_LOG_DEBUG, "%s(%d) ret = %d\n", __FUNCTION__, __LINE__,
               ret);
        if (ret < 0)
            goto error;
        i++;
    }

    ret = ff_formats_ref(in_fmts, &ctx->inputs[0]->outcfg.formats);
    av_log(ctx, AV_LOG_DEBUG, "%s(%d) ret = %d\n", __FUNCTION__, __LINE__, ret);
    if (ret < 0)
        goto error;

    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);

    return 0;

error:
    av_log(ctx, AV_LOG_ERROR, "%s failed\n", __FUNCTION__);

    if (constraints)
        av_hwframe_constraints_free(&constraints);
    if (hwconfig)
        av_freep(&hwconfig);

    return ret;
}

static av_cold int init(AVFilterContext *ctx)
{
    VpeGC620Context *s = ctx->priv;
    int i, ret;

    if (s->x > 50 / s->y) {
        av_log(ctx, AV_LOG_ERROR, "gc620 size %dx%d is insane.\n", s->x, s->y);
        return AVERROR(EINVAL);
    }

    if (!strcmp(s->p_type_str, "none"))
        return AVERROR(EINVAL);

    if (!strcmp(s->p_type_str, "blend"))
        s->p_type = VPE_PROCESSOR_BLEND;

    if (!strcmp(s->p_type_str, "tile"))
        s->p_type = VPE_PROCESSOR_TILE;

    if (!strcmp(s->p_type_str, "overlay"))
        s->p_type = VPE_PROCESSOR_OVERLAY;

    if (!strcmp(s->p_type_str, "drawbox"))
        s->p_type = VPE_PROCESSOR_DRAWBOX;

    /* processor_type of gc620 should be set in a continuous range */
    if (s->p_type < VPE_PROCESSOR_BLEND || s->p_type > VPE_PROCESSOR_DRAWBOX) {
        av_log(ctx, AV_LOG_ERROR, "gc620 processor type %d is out of range.\n",
               s->p_type);
        return AVERROR(EINVAL);
    }

    if (s->p_type == VPE_PROCESSOR_DRAWBOX) {
        if (av_parse_color(s->rgb_color, s->color_str, -1, ctx) < 0)
            return AVERROR(EINVAL);
        if (s->nb_inputs != 1)
            return AVERROR(EINVAL);
    }

    if (s->p_type == VPE_PROCESSOR_OVERLAY) {
        if (s->nb_inputs != 2)
            return AVERROR(EINVAL);
    }

    if (s->p_type == VPE_PROCESSOR_BLEND) {
        if (s->nb_inputs != 3)
            return AVERROR(EINVAL);
    }

    av_log(ctx, AV_LOG_DEBUG, "grid x=%d, y=%d.\n", s->x, s->y);

    s->frames = av_calloc(s->nb_inputs, sizeof(*s->frames));
    if (!s->frames)
        return AVERROR(ENOMEM);

    for (i = 0; i < s->nb_inputs; i++) {
        AVFilterPad pad = { 0 };

        pad.type = AVMEDIA_TYPE_VIDEO;
        pad.name = av_asprintf("input%d", i);
        if (!pad.name)
            return AVERROR(ENOMEM);

        if ((ret = ff_append_inpad_free_name(ctx, &pad)) < 0) {
            av_freep(&pad.name);
            return ret;
        }
    }

    return 0;
}

static int process_frame(FFFrameSync *fs)
{
    AVFilterContext *ctx = fs->parent;
    AVFilterLink *outlink = ctx->outputs[0];
    VpeGC620Context *s = fs->opaque;
    AVFrame **in = s->frames;
    AVFrame *out;
    VpiFrm *pic_in, *pic_out, *pic_temp, *pic;
    int i, ret;

    for (i = 0; i < s->nb_inputs; i++) {
        if ((ret = ff_framesync_get_frame(&s->fs, i, &in[i], 0)) < 0)
            return ret;

        /* set vpi input frames */
        pic_temp = (VpiFrm *)(in[i]->data[0]);
        if (!pic_temp)
            return AVERROR_INVALIDDATA;

        if (i == 0) {
            pic_in = pic_temp;
            pic    = pic_in;
        } else {
            pic->next = pic_temp;
            pic       = pic->next;
        }
    }

    /* need to get new video buffer */
    out = ff_get_video_buffer(outlink, outlink->w, outlink->h);
    if (!out)
        return AVERROR(ENOMEM);

    out->pts = av_rescale_q(s->fs.pts, s->fs.time_base, outlink->time_base);
    out->sample_aspect_ratio = outlink->sample_aspect_ratio;

    pic_out = (VpiFrm *)(out->data[0]);
    if (!pic_out)
        return AVERROR_INVALIDDATA;

    ret = vpi_processor_process_frame(s->vpe_gc620_handle, pic_in, pic_out);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "failed to process frames in@%p, out@%p, ret=%s\n",
               pic_in, pic_out, vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    return ff_filter_frame(outlink, out);
}

static int config_output(AVFilterLink *outlink)
{
    AVFilterContext *ctx = outlink->src;
    VpeGC620Context *s = ctx->priv;
    AVRational frame_rate = ctx->inputs[0]->frame_rate;
    AVRational sar = ctx->inputs[0]->sample_aspect_ratio;
    int height = ctx->inputs[0]->h;
    int width = ctx->inputs[0]->w;
    FFFrameSyncIn *in;
    int i, ret;
    VpiProcInitParams *params = &s->vpe_gc620_params;

    s->desc = av_pix_fmt_desc_get(outlink->format);
    if (!s->desc)
        return AVERROR_BUG;

    outlink->w          = width;
    outlink->h          = height;
    outlink->frame_rate = frame_rate;
    outlink->sample_aspect_ratio = sar;

    for (i = 1; i < s->nb_inputs; i++) {
        AVFilterLink *inlink = ctx->inputs[i];
        if (outlink->frame_rate.num != inlink->frame_rate.num ||
            outlink->frame_rate.den != inlink->frame_rate.den) {
            av_log(ctx, AV_LOG_VERBOSE,
                    "Video inputs have different frame rates, output will be VFR\n");
            outlink->frame_rate = av_make_q(1, 0);
            break;
        }
    }

    if ((ret = ff_framesync_init(&s->fs, ctx, s->nb_inputs)) < 0)
        return ret;

    in = s->fs.in;
    s->fs.opaque = s;
    s->fs.on_event = process_frame;

    for (i = 0; i < s->nb_inputs; i++) {
        AVFilterLink *inlink = ctx->inputs[i];

        in[i].time_base = inlink->time_base;
        in[i].sync   = 1;
        in[i].before = EXT_STOP;
        in[i].after  = s->shortest ? EXT_STOP : EXT_INFINITY;
    }

    if(ff_framesync_configure(&s->fs) < 0) {
        av_log(ctx, AV_LOG_ERROR, "failed to configure ff_framesync.\n");
        return AVERROR(EINVAL);
    }

    outlink->time_base = s->fs.time_base;

    /* init vpe processor handle */
    ret = vpe_init_hwctx(ctx);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "Vpe init hwctx failed.\n");
        return AVERROR_EXTERNAL;
    }

    outlink->hw_frames_ctx = av_buffer_ref(s->hw_frame_ctx);
    if (!outlink->hw_frames_ctx) {
        av_log(ctx, AV_LOG_ERROR, "outlink hw frame context get failed.\n");
        return AVERROR(ENOMEM);
    }

    /* get params */
    if (ctx->inputs[0]->format == AV_PIX_FMT_VPE) {
        AVHWFramesContext *fmctx =
            (AVHWFramesContext *)(ctx->inputs[0]->hw_frames_ctx->data);
        params->input_sw_format = av_get_pix_fmt_name(fmctx->sw_format);
        params->input_format    = av_get_pix_fmt_name(ctx->inputs[0]->format);
        av_log(ctx, AV_LOG_DEBUG,
               "%s(%d), fmctx->sw_format=%d, params->input_sw_format=%s.\n",
               __func__, __LINE__, fmctx->sw_format, params->input_sw_format);
    } else {
        av_log(ctx, AV_LOG_ERROR, "%s: Format: %s not supported.\n", __func__,
               av_get_pix_fmt_name(ctx->inputs[0]->format));
        return AVERROR(EINVAL);
    }

    if (s->overlay_x > width || s->overlay_y > height) {
        av_log(ctx, AV_LOG_ERROR,
               "overlay position is out of input boundary.\n");
        return AVERROR(EINVAL);
    }

    params->processor_type = s->p_type;
    params->core_id        = s->core_id;
    params->grid_x         = s->x;
    params->grid_y         = s->y;
    params->nb_frames      = s->nb_inputs;
    params->width          = width;
    params->height         = height;
    params->x_position     = s->overlay_x;
    params->y_position     = s->overlay_y;
    params->box_thickness  = s->thickness;
    params->box_color = ((((s->rgb_color[A]) & 0xFF) << 24) | (((s->rgb_color[R])&0xFF) << 16) |
	                    (((s->rgb_color[G])&0xFF) << 8) | ((s->rgb_color[B])&0xFF));

    av_log(ctx, AV_LOG_DEBUG,
        "params: grid_x=%d, grid_y=%d, nb_frames=%d, width=%d, height=%d.\n",
        params->grid_x, params->grid_y, params->nb_frames, params->width,
        params->height);

    ret = vpi_processor_init(params, &s->vpe_gc620_handle);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "%s(%d): GC620 processor init failed: %s!\n",
               __func__, __LINE__, vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    ret = vpi_processor_open(s->vpe_gc620_handle);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "%s(%d): GC620 processor open failed: %s.\n",
               __func__, __LINE__, vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }
    return ret;
}

static av_cold void uninit(AVFilterContext *ctx)
{
    VpeGC620Context *s = ctx->priv;

    ff_framesync_uninit(&s->fs);
    av_freep(&s->frames);

    av_buffer_unref(&s->hwdevice);
    av_buffer_unref(&s->hw_frame_ctx);

    if (s->vpe_gc620_handle)
        vpi_processor_close(s->vpe_gc620_handle);
}

static int activate(AVFilterContext *ctx)
{
    VpeGC620Context *s = ctx->priv;
    return ff_framesync_activate(&s->fs);
}

#define OFFSET(x) offsetof(VpeGC620Context, x)
#define FLAGS AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_FILTERING_PARAM

static const AVFilterPad outputs[] = {
    {
        .name          = "default",
        .type          = AVMEDIA_TYPE_VIDEO,
        .config_props  = config_output,
    },
};

static const AVOption gc620_vpe_options[] = {
    { "layout", "set grid size (for tiling processor)", OFFSET(x), AV_OPT_TYPE_IMAGE_SIZE, {.str="2x2"}, 0, 0, .flags = FLAGS},
    { "inputs", "set number of inputs", OFFSET(nb_inputs), AV_OPT_TYPE_INT, {.i64=2}, 1, INT_MAX, .flags = FLAGS },
    { "processor",  "set processor type", OFFSET(p_type_str), AV_OPT_TYPE_STRING, {.str = "none"}, .flags = FLAGS },
    { "core_id", "set core to run", OFFSET(core_id), AV_OPT_TYPE_INT, {.i64=0}, 0, 1, .flags = FLAGS},
    { "shortest", "force termination when the shortest input terminates", OFFSET(shortest), AV_OPT_TYPE_BOOL, {.i64=0}, 0, 1, .flags = FLAGS },
    { "color", "set color of the box only (for drawbox processor)", OFFSET(color_str), AV_OPT_TYPE_STRING, { .str = "black" }, CHAR_MIN, CHAR_MAX, .flags = FLAGS },
    { "thickness", "set the box thickness (for drawbox processor)", OFFSET(thickness), AV_OPT_TYPE_INT, { .i64=2 }, 2, INT_MAX, .flags = FLAGS },
    { "x", "Overlay x_position", OFFSET(overlay_x), AV_OPT_TYPE_INT, { .i64=0 }, 0, INT_MAX, .flags = FLAGS },
    { "y", "Overlay y_position", OFFSET(overlay_y), AV_OPT_TYPE_INT, { .i64=0 }, 0, INT_MAX, .flags = FLAGS },
    { NULL },
};

AVFILTER_DEFINE_CLASS(gc620_vpe);

const AVFilter ff_vf_gc620_vpe = {
    .name          = "2d_ama",
    .description   = NULL_IF_CONFIG_SMALL("2D Engine: blend drawbox tiling overlay."),
    .priv_size     = sizeof(VpeGC620Context),
    .priv_class    = &gc620_vpe_class,
    FILTER_OUTPUTS(outputs),
    FILTER_QUERY_FUNC(query_formats),
    .init          = init,
    .uninit        = uninit,
    .activate      = activate,
    .flags         = AVFILTER_FLAG_DYNAMIC_INPUTS | AVFILTER_FLAG_SLICE_THREADS,
    .flags_internal = FF_FILTER_FLAG_HWFRAME_AWARE,
};
